import unicodedata

import cbsodata
import pandas as pd
from django.conf import settings
from django.utils.translation import ugettext as _
from sqlalchemy import create_engine

from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.core.util.controllers import AbstractController
from iqportal.iqube.core.controllers import IQubeController, DataSourceController

# monkey patch below is a workaround for older pandas version that can't read 'nan' data
# http://stackoverflow.com/questions/23353732/python-pandas-write-to-sql-with-nan-values
from iqportal.iqube.core.models import DataSourceCube

if tuple([int(x) for x in pd.__version__.split('.')]) < (0, 15, 0):
    import numpy as np
    from pandas.io import sql
    def _write_mysql(frame, table, names, cur):
        bracketed_names = ['`' + column + '`' for column in names]
        col_names = ','.join(bracketed_names)
        wildcards = ','.join([r'%s'] * len(names))
        insert_query = "INSERT INTO %s (%s) VALUES (%s)" % (
            table, col_names, wildcards)
        data = [[None if type(y) == float and np.isnan(y) else y for y in x] for x in frame.values]
        cur.executemany(insert_query, data)
    sql._write_mysql = _write_mysql


class CBSController(AbstractController):
    def __init__(self, *args, **kwargs):
        super(CBSController, self).__init__(*args, **kwargs)
        # connection to cbs database
        self.engine = create_engine(settings.DB_CONNSTR_CBS)

    def import_data(self, cbs_id, force_update=False):
        """
        loads data into dataframe, imports data in dataframe into db
        updates data in db if force_update = 1
        """

        if cbs_id in (None, "", ):
            raise PublicUserException(_("Please select a datasource"))


        # assign permissions
        self.assert_permissions('manage_cbs_data')

        # convert the result into a pandas DataFrame, prints 'Retrieving data from table cbs_id Done!
        data = pd.DataFrame(cbsodata.get_data(cbs_id))

        # checks if the amount of columns is not greater than 1500,
        # it is decided that tables with more than 200 columns are not the way to go
        if len(data.columns)>200:
            raise PublicUserException("This CBS datasource contains too many columns to be imported")

        exists = self.engine.dialect.has_table(self.engine, cbs_id)

        if exists and not force_update:
            raise PublicUserException("CBS datasource %s has already been imported" % cbs_id)


        # makes table and inserts data if table not exists
        if not exists:
            data.to_sql(cbs_id, self.engine.raw_connection(), flavor='mysql', if_exists="replace")

        from iqportal.datasources.cbs.iqube_sync import CBSProject
        controller = IQubeController()
        project = CBSProject(self.sa_session, controller)
        project.execute()

    def get_tables(self):
        """
        returns a list of tablenames from the CBS database

        :return:
        """

        table_names = self.engine.execute("SHOW tables").fetchall()

        table_names = [table[0] for table in table_names]
        return table_names

    def get_cube_xml(self, cbs_id):
        """
        Creates a cube from input

        :param cbs_id:
        :return:
        """

        # list with dictionaries containing dataproperties
        metadata = cbsodata.get_meta('%s' % cbs_id, 'DataProperties')

        # list of for which are assigned to TopicGroups
        topic_group = [d for d in metadata if d['Type'] == 'TopicGroup']

        # list with dictionaries for dimensions
        dimensions = [d for d in metadata if d['Type'] in ('Dimension', 'TimeDimension',)]

        # list with dictionaries for measures
        measures = [d for d in metadata if d['Type'] in ('Topic',)]

        # Take care of the UTF-8 issues with the Mondrian schema
        for dim in dimensions:
            dim['Title'] = unicodedata.normalize('NFKD', dim['Title']).encode('ascii','ignore')


        for m in measures:

            # include title of topicgroup, if topigroup in metadata.
            if len(topic_group) > 0:
                topics = [tg['Title'] for tg in topic_group if tg['ID'] == m['ParentID']]
                if len(topics)>0:
                    m['Title'] = '%s (%s)' % (m['Title'], topics.pop())

            m['Title'] = unicodedata.normalize('NFKD', m['Title']).encode('ascii', 'ignore')


        dimension = """
                <Dimension name="%(Title)s">
                    <Hierarchy hasAll="true" allMemberName="All %(Title)s">
                        <Level name="%(Title)s" column="%(Key)s" />
                    </Hierarchy>
                </Dimension>
        """

        measure = """
                <Measure name="%(Title)s" column="%(Key)s" aggregator="avg" datatype="Numeric"/>
        """

        # dimensions % d loops over the list containing dictionaries, and writes them down.
        # same for measures
        cube = """
            <Cube name="%(name)s">

                <Table name="%(cbs_id)s"/>
                    %(dimensions)s
                    %(measures)s

            </Cube>
        """ % {
            'name': cbs_id,
            'cbs_id': cbs_id,
            'dimensions': ''.join([dimension % d for d in dimensions]),
            'measures': ''.join([measure % d for d in measures])
        }

        return cube

    def get_datasource(self):
        """
        testing existence of the datasource is done in test_import
        :return:
        """

        self.dsc = DataSourceController(self.sa_session, account_id=self.account_id)
        return self.dsc.get_datasource_by('name', '%s.CBSProject' % settings.SETTINGS_MODULE).pop()

    def get_all_tables(self):
        """
        Filters tables which are not imported.
        Returns a dictionary of ID, Tablename for tables which are not imported

        :return:
        """

        # extract tables in CBS db
        imported_tables = self.get_tables()

        #extract all tables for cbsodata
        tables = cbsodata.get_table_list()

        #if table in db
        return [{'value': d['Identifier'], 'label': d['ShortTitle']} for d in tables if d['Identifier'] not in imported_tables]

    def get_all_imported_cubes(self):
        """
        returns names of all imported tables
        :return:
        """

        # extract tables in CBS db of basesoft
        imported_tables = self.get_tables()

        # extract all tables for cbs website
        tables = cbsodata.get_table_list()

        rv = []
        for d in tables:

            cube_name = d['Identifier']

            if not cube_name in imported_tables:
                continue

            cube = self.sa_session.query(DataSourceCube).filter(DataSourceCube.name == cube_name).first()
            if len(self.get_privileged_object_by_id(DataSourceCube, cube.id)) == 0:
                continue

            rv.append({'value': cube_name, 'label': d['ShortTitle']})

        return rv



    def get_cubes(self):
        """
        returns all cubes from the CBSProject, which is below the 'mondrian_project' column in tblDataSource

        :param cbs_id:
        :return:
        """

        datasource = self.get_datasource()
        return self.dsc.get_cubes(datasource)

    def delete_table(self, cbs_id):
        """
        remove table from db if exists, if not pass
        :return:
        """

        self.assert_any_permissions('manage_cbs_data')
        self.engine.execute("DROP TABLE IF EXISTS %s" % (cbs_id))
