import json
from decimal import Decimal

import requests
from django.conf import settings
from django.utils.translation import gettext as _
from sqlalchemy import create_engine

from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.core.util.controllers import AbstractController, lazy_assert
from iqportal.datasources.oxford.api_client import Client as OxfordClient
from iqportal.iqube.core.controllers import IQubeController, DataSourceController, ItemResultController
from iqportal.iqube.core.models import DataSourceCube

BASE_URL = 'https://services.oxfordeconomics.com'

PERIOD_Q1 = 1
PERIOD_Q2 = 2
PERIOD_Q3 = 3
PERIOD_Q4 = 4
PERIOD_ANNUAL = 5

# values with more than below decimals get rounded
MAX_DECIMALS = 2


class OxfordController(AbstractController):
    def __init__(self, *args, **kwargs):
        super(OxfordController, self).__init__(*args, **kwargs)

        # connection to cbs database
        self.oxford_db = create_engine(settings.DB_CONNSTR_OXFORD)

        self.client = OxfordClient(settings.DATASOURCE_OXFORD_API_KEY)

    def get_imported_databanks(self):
        return self.oxford_db.execute("SELECT `id`, `code`, `title` FROM oxford_databank ORDER BY `id`")

    def get_cube_names(self):
        rv = []
        for db in self.get_imported_databanks():
            cube_name = 'Oxford%s' % db['code']
            cube = self.sa_session.query(DataSourceCube).filter(DataSourceCube.name == cube_name).first()
            if len(self.get_privileged_object_by_id(DataSourceCube, cube.id)) == 0:
                continue
            rv.append({'value': db['id'], 'label': db['title']})

        return rv

    def get_available_databanks(self):

        rv = []

        for databank in self.client.get_databanks():

            if not databank['HasAccess']:
                continue

            rv.append({
                'code': databank['DatabankCode'],
                'name': databank['Name']
            })

        return rv

    def import_data(self, databank_str, force_update=False):
        selection = self._build_selection(databank_str)

        # download saves to a temp file called 'oxford.json'
        data = self._download_databank(selection)

        # error checking
        if type(data) == dict:
            assert data['Message'] != 'BadRequest', data

        self._import_tables(databank_str, data)

        self._sync_cubes()

        self.logger.info('imported databank: %s' % databank_str)

    def _download_url(self, base_url, page, page_size):
        url = base_url + '/api/download?includemetadata=true'
        return url + '&page={1}+&pagesize={2}'.format(base_url, page, page_size)

    def _download_databank(self, selection):
        headers = {'Accept': 'application/json',
                   'Api-Key': settings.DATASOURCE_OXFORD_API_KEY,
                   'Content-Type': 'application/json; charset=utf8'}

        page = 0
        page_size = 5000  # 20000 is max
        data = None

        while True:
            # note: _download_url returns a link of the form
            # ../api/download?includemetadata=true&page={page}+&page_size={page_size}
            response = requests.post(self._download_url(BASE_URL, page, page_size),
                                     headers=headers,
                                     data=json.dumps(selection))

            new_data = response.json()
            page += 1

            if data is None:
                data = new_data
            else:
                # data = data.append(new_data)
                data += new_data

            if len(new_data) != page_size:
                break

        return data

    def _build_selection(self, databank_code):
        lazy_assert(databank_code, str)

        # from http://www.oxfordeconomics.com/techlabs/api-guide-for-the-global-data-workstation#databank-download-example

        selection = {
            'DatabankCode': databank_code,
            'Frequency': 'Both',  # or Annual | Quarterly | Both
            'GroupingMode': 'false',
            'IndicatorSortOrder': 'AlphabeticalOrder',
            'IsTemporarySelection': 'true',
            'ListingType': 'Private',
            'LocationSortOrder': 'AlphabeticalOrder',
            'Order': 'IndicatorLocation',
            'Precision': 1,
            'Sequence': 'EarliestToLatest',
            'StackedQuarters': 'false',
            'StartYear': 2010,
            'EndYear': 2045,

            # note: the fields below have been assigned empty lists
            'Regions': [
            ],
            'Variables': [
            ]
        }

        return selection

    def _import_tables(self, databank_code, data):

        databank_id = None
        q = self.oxford_db.execute("SELECT `id` FROM oxford_databank WHERE code=%s", databank_code)
        for row in q:
            databank_id = row[0]

        if not databank_id:
            databank_title = self.client.get_databanks(databank_code)['Name']
            q = self.oxford_db.execute("INSERT INTO oxford_databank (code, title) VALUES (%s, %s) ", databank_code,
                                       databank_title)
            databank_id = q.lastrowid

        assert databank_id > 0

        # stats
        self._max_decimals = 0
        self._max_nr = 0
        self._facts = 0

        # trash old data for this databank
        self.oxford_db.execute("DELETE FROM oxford_fact WHERE databank_id=%s", databank_id)
        # self.oxford_db.execute("DELETE FROM oxford_indicator WHERE databank_id=%s", databank_id)

        # import new indicators
        self._import_indicators(databank_code, databank_id)

        for line in data:

            q = self.oxford_db.execute("""SELECT `id` FROM oxford_indicator 
                                       WHERE databank_id=%s AND code=%s AND product_type_code=%s""",
                                       (databank_id, line['VariableCode'], line['ProductTypeCode']))
            indicator_ids = []
            for row in q:
                indicator_ids.append(int(row['id']))
            assert len(indicator_ids) > 0

            # insert/get location id

            location_code = line['LocationCode']
            location_title = line['Metadata']['Location']

            location_id = None
            q = self.oxford_db.execute("SELECT `id` FROM oxford_location WHERE code=%s", (location_code))
            for row in q:
                location_id = row[0]

            if not location_id:
                q = self.oxford_db.execute("INSERT INTO oxford_location (code, title) VALUES (%s, %s) ", location_code,
                                           location_title)
                location_id = q.lastrowid
            assert location_id > 0

            for indicator_id in indicator_ids:

                # insert annual data
                if len(line['AnnualData']):
                    for year, value in line['AnnualData'].iteritems():
                        self._insert_fact(databank_id, location_id, int(year), PERIOD_ANNUAL, indicator_id, value)

                # quarterly
                if len(line['QuarterlyData']):
                    for yearqq, value in line['QuarterlyData'].iteritems():

                        try:
                            year = int(yearqq[:4])
                            quarter = int(yearqq[4:])
                        except ValueError:
                            raise Exception('invalid: %s' % yearqq)
                        assert quarter in (PERIOD_Q1, PERIOD_Q2, PERIOD_Q3, PERIOD_Q4)

                        self._insert_fact(databank_id, location_id, year, quarter, indicator_id, value)

        self.logger.info('finished importing databank_code: %s. number of facts: %s' % (databank_code, self._facts))

    def _insert_fact(self, databank_id, location_id, year, period, indicator_id, value):

        if value is None:
            return

        assert 2010 <= year <= 2045, 'invalid year: %s' % year

        try:
            decimals = abs(Decimal(str(value)).as_tuple().exponent)
        except IndexError:
            raise Exception('invalid decimal: %s' % value)

        if decimals > MAX_DECIMALS:
            value = round(Decimal(value), MAX_DECIMALS)

        if decimals > self._max_decimals:
            self._max_decimals = decimals

        nrs = len(str(value).split('.')[0])
        if nrs > self._max_nr:
            max_nr = nrs

        q = self.oxford_db.execute("""
            INSERT INTO oxford_fact (databank_id, location_id, year, period, indicator_id, value)
            VALUES (%s, %s, %s, %s, %s, %s) """, databank_id, location_id, year, period,
                                   indicator_id, value)

        self._facts = self._facts + 1

    def _sync_cubes(self):

        from iqportal.datasources.oxford.iqube_sync import OxfordProject
        controller = IQubeController()
        project = OxfordProject(self.sa_session, controller)
        project.execute()

    def get_datasource(self):
        self.dsc = DataSourceController(self.sa_session, account_id=self.account_id)
        return self.dsc.get_datasource_by('name', '%s.OxfordProject' % settings.SETTINGS_MODULE).pop()

    def make_configurable_widget_config(self, databank_id):
        """
        """
        if databank_id in (None, "",):
            raise PublicUserException(_("Please select a cube."))

        datasource_id = self.get_datasource().id

        q = self.oxford_db.execute("SELECT `code` FROM oxford_databank WHERE `id`=%s", databank_id)
        for row in q:
            code = row['code']
        cube_name = 'Oxford%s' % code

        c = ItemResultController(self.sa_session, account_id=self.account_id)
        rv = c.make_configurable_widget_config(datasource_id, cube_name)

        # make all dimensions required
        for dimension in rv['config']['dimensions']:
            dimension['available_as_slicer'] = True
            dimension['show_named_sets'] = False
            dimension['has_select_as_group'] = False
            dimension['required'] = True
            dimension['fancy_select_module'] = 'iq_fancyselect'
            dimension['children'][0]['has_children_available'] = False
            dimension['children'][0]['has_all_member_available'] = False


        # point to new settings object
        rv['config']['filter_panel_module'] = 'oxford_filter_module'

        return rv

    def _get_databank_id(self, databank_code):
        databank_code
        databank_id = None
        q = self.oxford_db.execute("SELECT `id` FROM oxford_databank WHERE code=%s", databank_code)
        for row in q:
            databank_id = row[0]
        lazy_assert(databank_id, int)
        return databank_id

    def get_filter_tree(self, cube_name):
        databank_str = cube_name[len('Oxford'):]  # strip 'Oxford' prefix
        databank_id = self._get_databank_id(databank_str)

        rv = []

        def _get_children(node):
            q = self.oxford_db.execute("""
                SELECT * FROM oxford_indicator 
                WHERE `databank_id`=%s 
                AND `parent_id` <=> %s
                ORDER BY display_order
            """, (databank_id, node['id']))
            for row in q:
                child_node = {
                    'id': row['id'],
                    'label': row['name'],
                    'children': []
                }
                _get_children(child_node)

                node['children'].append(child_node)

        root = {
            'id': None,
            'label': 'Root',
            'children': []
        }
        _get_children(root)

        return root

    def _import_indicators(self, databank_str, databank_id):

        client = OxfordClient(settings.DATASOURCE_OXFORD_API_KEY)
        tree = client.get_indicators_tree(databank_str)

        for child in tree:
            self._import_indicator(databank_id, None, child)

            # TODO: implement DELETE for indicators that are in the DB but not here anymore

    def _import_indicator(self, databank_id, parent_id, entry):

        code = entry['Code']
        product_type_code = entry['ProductTypeCode']
        display_order = entry['DisplayOrder']
        name = entry['Name']

        if not entry['HasAccess']:
            return

        indicator_id = None

        # see if indicator already exists
        q = self.oxford_db.execute("""
            SELECT `id` FROM oxford_indicator 
            WHERE databank_id=%s AND parent_id <=> %s AND code <=> %s AND product_type_code <=> %s
            AND display_order=%s AND name=%s
            """, databank_id, parent_id, code, product_type_code, display_order, name)

        for row in q:
            indicator_id = row[0]

        if indicator_id is None:
            q = self.oxford_db.execute("""
                INSERT INTO oxford_indicator (databank_id, parent_id, code, product_type_code, display_order, name) 
                VALUES (%s, %s, %s, %s, %s, %s) """, databank_id, parent_id, code, product_type_code, display_order,
                                       name)

            indicator_id = q.lastrowid

        assert indicator_id > 0

        if entry.has_key('Children'):
            for child in entry['Children']:
                self._import_indicator(databank_id, indicator_id, child)
